#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import uuid as sys_uuid

import pytest

from restalchemy.dm import models as ra_models
from restalchemy.dm import properties
from restalchemy.dm import types as ra_types
from restalchemy.api import applications
from restalchemy.api import middlewares
from restalchemy.api import controllers
from restalchemy.api import routes
from restalchemy.api.middlewares import contexts as context_mw
from restalchemy.api.middlewares import logging as logging_mw
from restalchemy.api.middlewares import errors as errors_mw
from restalchemy.openapi import structures as openapi_structures
from restalchemy.openapi import engines as openapi_engines

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.orch_api import routes as orch_routes
from gcl_sdk.agents.universal.status_api import routes as status_routes


FIRST_MIGRATION = "0000-init-events-table-2cfd220e.py"


class FooResource(ra_models.ModelWithUUID, models.ResourceMixin):
    name = properties.property(
        ra_types.String(max_length=64), default="foo-name"
    )
    project_id = properties.property(
        ra_types.UUID(), default=lambda: sys_uuid.uuid4()
    )


class FooTargetResource(ra_models.ModelWithUUID, models.TargetResourceMixin):
    name = properties.property(
        ra_types.String(max_length=64), default="foo-name"
    )
    project_id = properties.property(
        ra_types.UUID(), default=lambda: sys_uuid.uuid4()
    )


def get_openapi_engine():
    openapi_engine = openapi_engines.OpenApiEngine(
        info=openapi_structures.OpenApiInfo(
            title="Test API",
            version="v1",
            description="OpenAPI - Test API",
        ),
        paths=openapi_structures.OpenApiPaths(),
        components=openapi_structures.OpenApiComponents(),
    )
    return openapi_engine


@pytest.fixture(scope="module")
def orch_api_wsgi_app():
    class OrchApiApp(routes.RootRoute):
        pass

    class ApiEndpointController(controllers.RoutesListController):
        __TARGET_PATH__ = "/v1/"

    class ApiEndpointRoute(routes.Route):
        __controller__ = ApiEndpointController
        __allow_methods__ = [routes.FILTER]

        agents = routes.route(orch_routes.UniversalAgentsRoute)

    setattr(
        OrchApiApp,
        "v1",
        routes.route(ApiEndpointRoute),
    )

    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=OrchApiApp,
            openapi_engine=get_openapi_engine(),
        ),
        [
            context_mw.ContextMiddleware,
            errors_mw.ErrorsHandlerMiddleware,
            logging_mw.LoggingMiddleware,
        ],
    )


@pytest.fixture(scope="module")
def status_api_wsgi_app():
    class StatusApiApp(routes.RootRoute):
        pass

    class ApiEndpointController(controllers.RoutesListController):
        __TARGET_PATH__ = "/v1/"

    class ApiEndpointRoute(routes.Route):
        __controller__ = ApiEndpointController
        __allow_methods__ = [routes.FILTER]

        agents = routes.route(status_routes.UniversalAgentsRoute)
        kind = routes.route(status_routes.KindRoute)

    setattr(
        StatusApiApp,
        "v1",
        routes.route(ApiEndpointRoute),
    )

    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=StatusApiApp,
            openapi_engine=get_openapi_engine(),
        ),
        [
            context_mw.ContextMiddleware,
            errors_mw.ErrorsHandlerMiddleware,
            logging_mw.LoggingMiddleware,
        ],
    )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import typing as tp
import importlib_metadata


EVENT_PAYLOADS_GROUP = "gcl_sdk_event_payloads"


def load_event_payload_map() -> dict:
    event_payload_map = {
        ep.name: ep.load()
        for ep in importlib_metadata.entry_points(
            group=EVENT_PAYLOADS_GROUP,
        )
    }
    return event_payload_map


def load_from_entry_point(group: str, name: str) -> tp.Any:
    """Load class from entry points."""
    for ep in importlib_metadata.entry_points(group=group):
        if ep.name == name:
            return ep.load()

    raise RuntimeError(f"No class '{name}' found in entry points {group}")



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc
import typing as tp

from gcl_sdk.agents.universal.dm import models


class AbstractAgentStorage(abc.ABC):
    """Abstract agent storage."""

    @abc.abstractmethod
    def get(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Get the resource  item from the storage."""

    @abc.abstractmethod
    def create(
        self,
        resource: models.Resource,
        item: dict[str, tp.Any],
        force: bool = False,
    ) -> dict[str, tp.Any]:
        """Creates the resource item in the storage."""

    @abc.abstractmethod
    def update(
        self, resource: models.Resource, item: dict[str, tp.Any]
    ) -> dict[str, tp.Any]:
        """Update the resource item in the storage."""

    @abc.abstractmethod
    def list(self, kind: str) -> list[dict[str, tp.Any]]:
        """Lists all resource items by kind."""

    @abc.abstractmethod
    def delete(self, resource: models.Resource, force: bool = False) -> None:
        """Delete the resource item from the storage."""



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import json
import typing as tp
import uuid as sys_uuid

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.storage import base
from gcl_sdk.agents.universal.storage import exceptions as se


class FileAgentStorage(base.AbstractAgentStorage):
    """Rest API backend client."""

    DEFAULT_STORAGE_NAME = "ua_driver_storage.json"

    def __init__(self, work_dir: str, storage_name: str | None = None) -> None:
        self._storage_name = storage_name or self.DEFAULT_STORAGE_NAME
        self._storage_path = os.path.join(work_dir, self._storage_name)

    def _load(self, kind: str) -> list[dict[str, tp.Any]]:
        if not os.path.exists(self._storage_path):
            return []

        try:
            with open(self._storage_path) as f:
                return json.load(f)[kind]
        except KeyError:
            return []

    def _delete(self, kind: str, uuid: sys_uuid.UUID) -> None:
        """Remove the resource from the meta file."""
        if not os.path.exists(self._storage_path):
            return

        uuid = str(uuid)

        with open(self._storage_path, "r+") as f:
            data = json.load(f)
            data[kind] = [r for r in data[kind] if r["uuid"] != uuid]
            f.seek(0)
            f.truncate(0)
            json.dump(data, f, indent=2)

    def _add(self, kind: str, item: dict[str, tp.Any]) -> None:
        """Add the resource from the meta file."""
        # Create the directory if it doesn't exist
        os.makedirs(os.path.dirname(self._storage_path), exist_ok=True)

        # Create the file if it doesn't exist
        if not os.path.exists(self._storage_path):
            with open(self._storage_path, "w") as f:
                json.dump({kind: []}, f, indent=2)

        # Save the new meta object
        with open(self._storage_path, "r+") as f:
            data = json.load(f)
            if kind not in data:
                data[kind] = []
            data[kind].append(item)
            f.seek(0)
            f.truncate(0)
            json.dump(data, f, indent=2)

    def get(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Get the resource  item from the storage."""
        uuid = str(resource.uuid)

        for item in self._load(resource.kind):
            if item["uuid"] == uuid:
                return item

        raise se.ResourceNotFound(resource=resource)

    def create(
        self,
        resource: models.Resource,
        item: dict[str, tp.Any],
        force: bool = False,
    ) -> dict[str, tp.Any]:
        """Creates the resource item in the storage."""
        try:
            self.get(resource)
        except se.ResourceNotFound:
            # Desirable behavior, the resource should not exist
            pass
        else:
            if not force:
                raise se.ResourceAlreadyExists(resource)
            self._delete(resource.kind, resource.uuid)

        self._add(resource.kind, item)
        return item

    def update(
        self, resource: models.Resource, item: dict[str, tp.Any]
    ) -> dict[str, tp.Any]:
        """Update the resource item in the storage."""
        self.delete(resource)
        self.create(resource, item)
        return item

    def list(self, kind: str) -> list[dict[str, tp.Any]]:
        """Lists all resource items by kind."""
        return self._load(kind)

    def delete(self, resource: models.Resource, force: bool = False) -> None:
        """Delete the resource item from the storage."""
        try:
            self.get(resource)
        except se.ResourceNotFound:
            if not force:
                raise

        self._delete(resource.kind, resource.uuid)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import abc

from gcl_sdk.agents.universal.dm import models


class AbstractCapabilityDriver(abc.ABC):
    """Abstract driver for capabilities.

    The main purpose of this driver is to apply target
    capabilities to the DP.
    """

    @abc.abstractmethod
    def get_capabilities(self) -> list[str]:
        """Returns a list of capabilities supported by the driver."""

    @abc.abstractmethod
    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the data plane.
        """

    @abc.abstractmethod
    def create(self, resource: models.Resource) -> models.Resource:
        """Creates a resource."""

    @abc.abstractmethod
    def update(self, resource: models.Resource) -> models.Resource:
        """Update the resource.

        The simplest implementation. The driver should detect which
        fields were changed itself.
        """

    @abc.abstractmethod
    def list(self, capability: str) -> list[models.Resource]:
        """Lists all resources by capability."""

    @abc.abstractmethod
    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""


class AbstractFactDriver(abc.ABC):
    """Abstract driver for facts.

    The main purpose of this driver is to gather facts from the
    data plane. The simplest example is gathering network interfaces
    from nodes.
    """

    @abc.abstractmethod
    def get_facts(self) -> list[str]:
        """Returns a list of facts supported by the driver."""

    @abc.abstractmethod
    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the data plane.
        """

    @abc.abstractmethod
    def list(self, fact: str) -> list[models.Resource]:
        """Lists all resources by facts."""



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import uuid as sys_uuid

from gcl_sdk.agents.universal.drivers import base
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.clients.backend import base as client_base
from gcl_sdk.agents.universal.clients.backend import exceptions as client_exc
from gcl_sdk.agents.universal.storage import base as storage_base

LOG = logging.getLogger(__name__)


class DirectAgentDriver(base.AbstractCapabilityDriver):
    """Direct driver. Directly gets all data from the backend.

    The key feature of this driver it's able to get all
    data from the data plane without using any additional
    information like meta files.
    """

    def __init__(
        self,
        client: client_base.AbstractBackendClient,
        storage: storage_base.AbstractAgentStorage,
    ):
        super().__init__()
        self._client = client
        self._storage = storage

    def _create_storage_item(
        self, uuid: sys_uuid.UUID, target_fields: frozenset[str]
    ):
        return {"uuid": str(uuid), "target_fields": list(target_fields)}

    def _validate(self, resource: models.Resource) -> None:
        """Validate the resource."""
        if resource.kind not in self.get_capabilities():
            raise TypeError(f"Unsupported capability {resource.kind}")

    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the backend.
        """
        self._validate(resource)

        try:
            value = self._client.get(resource)
        except client_exc.ResourceNotFound:
            LOG.error("Unable to find resource on backend %s", resource.uuid)
            raise driver_exc.ResourceNotFound(resource=resource)

        # Figure out the target fields to correct hash calculation
        target_fields = frozenset(resource.value.keys())

        return resource.replace_value(value, target_fields)

    def list(self, capability: str) -> list[models.Resource]:
        """Lists all resources by capability."""
        if capability not in self.get_capabilities():
            raise TypeError(f"Unsupported capability {capability}")

        # Collect all resources in convient format
        resources = []
        storage_items = {i["uuid"]: i for i in self._storage.list(capability)}
        values = {i["uuid"]: i for i in self._client.list(capability)}

        # If storage item or value is missing, consider it as
        # a missing resource
        for uuid in storage_items.keys() & values.keys():
            value = values[uuid]
            item = storage_items[uuid]
            target_fields = frozenset(item["target_fields"])
            res = models.Resource.from_value(value, capability, target_fields)
            resources.append(res)

        return resources

    def create(self, resource: models.Resource) -> models.Resource:
        """Creates a resource."""
        self._validate(resource)

        # Figure out the target fields to correct hash calculation
        target_fields = frozenset(resource.value.keys())

        item = self._create_storage_item(resource.uuid, target_fields)

        try:
            value = self._client.create(resource)
            LOG.debug("Created resource: %s", resource.uuid)
        except client_exc.ResourceAlreadyExists:
            self._storage.create(resource, item, force=True)
            LOG.error("The resource already exists: %s", resource.uuid)
            raise driver_exc.ResourceAlreadyExists(resource=resource)

        self._storage.create(resource, item, force=True)
        return resource.replace_value(value, target_fields)

    def update(self, resource: models.Resource) -> models.Resource:
        """Update the resource."""
        self._validate(resource)

        try:
            value = self._client.update(resource)
            LOG.debug("Updated resource: %s", resource.uuid)
        except client_exc.ResourceNotFound:
            LOG.error("The resource does not exist: %s", resource.uuid)
            raise driver_exc.ResourceNotFound(resource=resource)

        # Figure out the target fields to correct hash calculation
        target_fields = frozenset(resource.value.keys())

        item = self._create_storage_item(resource.uuid, target_fields)
        self._storage.create(resource, item, force=True)

        return resource.replace_value(value, target_fields)

    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""
        self._validate(resource)

        try:
            self._client.delete(resource)
            LOG.debug("Deleted resource: %s", resource.uuid)
        except client_exc.ResourceNotFound:
            self._storage.delete(resource, force=True)
            LOG.debug("The resource is already deleted: %s", resource.uuid)

        self._storage.delete(resource, force=True)
